#!/usr/bin/python -d
import openbabel, sys, math, random
import numpy
from exceptions import Exception

'''
Initial hexagonal membrane bilayer simulation.

Alex M. Nesterenko

ERG Research Group
Department of Biophysics
Biology Faculty
Lomonosov Moscow State University

Please cite: 
  Nesterenko A.M., et al. Biochemistry (Moscow) Supplement Series A: Membrane and Cell Biology, 2011, 5(4), 370–378.
  doi: 10.1134/S1990747811050084
  
'''

try:
        cos30 = math.sqrt(3)/2
	def add(x, y): return x+y
	obC = openbabel.OBConversion()
	obC.SetInAndOutFormats("pdb", "pdb")

	mols = []
	verts = []
	probs = []

	# initializing structures
	if True:
	  nmol = openbabel.OBMol()
	  obC.ReadFile(nmol, "dppsnew.pdb")
	  mols.append( nmol )
	  verts.append( 56 )
	  probs.append( 0.6 )
	  nmol = openbabel.OBMol()
	  obC.ReadFile( nmol, "chlnew.pdb")
	  mols.append( nmol )
	  verts.append( 27 ) 
	  probs.append( 0 )
	  nmol = openbabel.OBMol()
	  obC.ReadFile( nmol, "dppc.pdb")
	  mols.append( nmol )
	  verts.append( 58 ) 
	  probs.append( 0.4 )

	# init values for grid
	n = 9
	m = 10
	step = 7.7
	zdist = 30.0
	outfile = "60s.pdb"
	numconformers = 1
        angle_confs = 10
	grid = 'hexagonal'

	ml_count = None
	# printing info
	for mol in mols:
	  print 'Using molecule: %s (%d atoms) ' % ( mol.GetResidue(0).GetName(), mol.NumAtoms() )

        # molecule rotation
	for i in range(0, len(mols)):
          if (probs[i] == 0):
              continue
	  mol = mols[i]
	  print 'Rotating: ',mol.GetResidue(0).GetName()
	  vert = verts[i]
	  a_vert = mol.GetAtom( vert )
	  far_atom = a_vert
	  for at in openbabel.OBMolAtomIter(mol):
	    if (at.GetIdx() == vert):
	      continue
	    if ( a_vert.GetDistance(at) > a_vert.GetDistance(far_atom) ):
	      far_atom = at
	  print 'Vertex atom = ',a_vert.GetIdx(),':',a_vert.GetType()
	  print 'Far atom    = ',far_atom.GetIdx(),':',far_atom.GetType()
	  mol.Translate(openbabel.vector3(-a_vert.x(), -a_vert.y(), -a_vert.z()))
	  angz = -math.atan(far_atom.y()/far_atom.x())
	  mol.Rotate(openbabel.double_array([math.cos(angz), -math.sin(angz), 0, math.sin(angz),math.cos(angz), 0, 0, 0, 1]))
	  angy = -math.atan(far_atom.x()/far_atom.z())
	  mol.Rotate(openbabel.double_array([math.cos(angy), 0, math.sin(angy), 0, 1, 0, -math.sin(angy), 0, math.cos(angy)]))
	  print 'Angles = ',angz*180/math.pi, angy*180/math.pi
	  mol.Center()
	  if (a_vert.z() < 0):
		mol.Rotate(openbabel.double_array([1, 0, 0,   0, -1, 0,  0, 0, -1]))

        # generate conformers
        mi = -1
	for mol in mols:
                mi += 1
                if (probs[mi] == 0):
                    continue
		for i in range(0,numconformers):
			nmol = openbabel.OBMol(mol)
			for torl in openbabel.OBMolTorsionIter(nmol):
				atlist = [nmol.GetAtom(torl[0]+1), nmol.GetAtom(torl[1]+1), nmol.GetAtom(torl[2]+1), nmol.GetAtom(torl[3]+1)]
				if (atlist[1].IsInRing() and atlist[2].IsInRing() ):
					continue
				tor = nmol.GetTorsion(atlist[0], atlist[1], atlist[2], atlist[3])
				tor = random.gauss(tor, 30)*math.pi/180
				nmol.SetTorsion(atlist[0], atlist[1], atlist[2], atlist[3], tor)
			mol.AddConformer(nmol.GetConformer(0))


	# generate shuffle sequence
	if (grid == 'square'):
		ml_count = n*m
        elif (grid == 'hexagonal'):
                ml_count = n*m

	lipids = [-1]*ml_count
	cur = 0
	for i in range(0,len(mols)):
		cnt = min([cur + int(probs[i]*ml_count), ml_count]) - cur
		lipids[cur:cur+cnt] = [i]*cnt
		cur += cnt
	if (cur < ml_count):
		lipids[ml_count-1] = 0

	random.shuffle(lipids)

	print 'Sequence after shuffling: '
	print lipids

	allmols = openbabel.OBMol()

	# square grid

	# first monolayer
	cm = 0
	centmass = []
	for i in range(0, n):
	  for j in range(0, m):
	    nmol = openbabel.OBMol(mols[lipids[cm]])
            if (grid == 'square'):
	        v3c = numpy.array([i*step, j*step, 0])
            elif (grid == 'hexagonal'):
                v3c = numpy.array([i*step+j*step/2, j*step*cos30, 0])
            print 'Disposition: ', i,j,v3c
            centmass.append(v3c)
	    nmol.GetResidue(0).SetNum(cm+1)
	    minconf = 0
            minangle = 0.0
	    maxdist = 0.0
	    needresids = []
	    for rrr in range(0, cm):
		    a__ = v3c - centmass[rrr]
		    d__ = math.sqrt(abs(numpy.dot(a__,a__)))
		    if ( d__ < 2*step ):
			    needresids.append(rrr+1)
#	    print needresids
            
            for k in range(0, angle_confs):
               if (k == 0):
                   aoo = 0.0
               else:
                   aoo = random.gauss(0, 180)/180*math.pi
               print aoo,':'
               nmol.Center()
               nmol.Rotate( openbabel.double_array([math.cos(aoo), -math.sin(aoo), 0, math.sin(aoo), math.cos(aoo), 0, 0, 0, 1]) )
               nmol.Translate(openbabel.vector3(v3c[0], v3c[1], v3c[2]))

               for conf in range(0, nmol.NumConformers()):
                      print conf,',',
                      nmol.SetConformer(conf)
                      minD = 999
                      for a0 in openbabel.OBMolAtomIter(allmols):
                              if ( not a0.GetResidue().GetNum() in needresids ):
                                      continue
                              for a1 in openbabel.OBMolAtomIter(nmol):
                                      dist = a0.GetDistance(a1)
                                      if ( minD > dist ):
                                              minD = dist
                      if (minD == 999): # some nan or inf ... (bad conformer)
                              continue
                      if (minD > maxdist):
                              minconf = conf
                              maxdist = minD
                              minangle = aoo

	    print '\nLipid #%4d  :  Selected conformer - %2d, angle - %8.3f = %8.3f :' % (cm, minconf, minangle/math.pi*180.0, maxdist), needresids
	    need_data = nmol.GetConformer(minconf)
            nmol.Center()
            nmol.Rotate( openbabel.double_array([math.cos(minangle),
              -math.sin(minangle), 0, math.sin(minangle), math.cos(minangle),
              0, 0, 0, 1]) )
            nmol.Translate(openbabel.vector3(v3c[0], v3c[1], v3c[2]))
	    nmol.SetConformer(0)
	    nmol.SetCoordinates(need_data)

	    allmols += nmol
	    cm += 1

	allmols.Center()
	# second monolayer
	try:
		sml = openbabel.OBMol(allmols)
		sml.Rotate(openbabel.double_array([1, 0, 0,   0, -1, 0,  0, 0, -1]))
		sml.Translate(openbabel.vector3(0, 0, -zdist))
		allmols += sml
	except Error:
		print 'Error during second monolayer adding'

	# saving..
	allmols.SetTitle('Membrane generator output')
	obC.WriteFile(allmols, outfile)

	# destructing
finally:
	print 'BYE-obc!'
